import logging

from senor_octopus.types import Stream
from slack_sdk import WebClient

_logger = logging.getLogger(__name__)


async def slack(
    stream: Stream,
    token: str,
    channel: str,
) -> None:
    """
    Send events as messages to a Slack channel.

    This sink can be used to send events to a Slack channel.
    The value of the event is sent as the message; its name
    is ignored.

    Parameters
    ----------
    stream
        The incoming stream of events
    token
        The authentication token for the bot
    channel
        The Slack channel ID
    """
    client = WebClient(token=token)
    async for event in stream:  # pragma: no cover
        client.chat_postMessage(channel=channel, text=str(event["value"]))



import logging

from jinja2 import Template
from senor_octopus.types import Stream

_logger = logging.getLogger(__name__)


async def jinja(stream: Stream, template: str) -> Stream:
    """
    Apply a Jinja2 template to events.

    This filter can be used to transform events according to
    a template, or filter them (templates that return nothing
    filter out the corresponding event).

    Parameters
    ----------
    stream
        The incoming stream of events
    template
        A Jinja2 template

    Yields
    ------
    Event
        Events filtered and/or transformed by the template
    """
    _logger.debug("Applying template to events")
    tmpl = Template(template)
    async for event in stream:  # pragma: no cover
        value = tmpl.render(event=event).strip()
        if value:
            yield {
                "timestamp": event["timestamp"],
                "name": event["name"],
                "value": value,
            }



import asyncio
import logging
from datetime import datetime
from datetime import timedelta
from datetime import timezone

from senor_octopus.types import Stream
from suntime import Sun

_logger = logging.getLogger(__name__)


async def sun(
    latitude: float,
    longitude: float,
    prefix: str = "hub.sun",
) -> Stream:
    """
    Send events on sunrise and sunset.

    This source will send a "sunrise" message on sunrise, and a
    "sunset" message on sunset.

    Parameters
    ----------
    latitude
        Latitude of the location, as a float (eg, -45.2)
    longitude
        Longitude of the location, as a float (eg, -200.1)

    Yields
    ------
    Event
        Sunrise and sunset events
    """
    _logger.info("Waiting for sunrise/sunset")
    info = Sun(latitude, longitude)

    while True:
        now = datetime.now(timezone.utc).astimezone()
        _logger.debug("Now is: %s", now)

        reference = now.date()

        next_sunrise = info.get_local_sunrise_time(reference)
        _logger.debug("Next sunrise: %s", next_sunrise)
        next_sunset = info.get_local_sunset_time(reference)
        if next_sunset < next_sunrise:
            next_sunset = info.get_local_sunset_time(reference + timedelta(days=1))
        _logger.debug("Next sunset: %s", next_sunset)

        if now < next_sunrise:
            _logger.debug("Waiting for today's sunrise")
            await asyncio.sleep((next_sunrise - now).total_seconds())
            yield {
                "timestamp": next_sunrise.astimezone(timezone.utc),
                "name": prefix,
                "value": "sunrise",
            }
        elif now < next_sunset:
            _logger.debug("Waiting for today's sunset")
            await asyncio.sleep((next_sunset - now).total_seconds())
            yield {
                "timestamp": next_sunset.astimezone(timezone.utc),
                "name": prefix,
                "value": "sunset",
            }
        else:
            _logger.debug("Waiting for tomorrow's sunrise")
            tomorrow = reference + timedelta(days=1)
            next_sunrise = info.get_local_sunrise_time(tomorrow)
            await asyncio.sleep((next_sunrise - now).total_seconds())
            yield {
                "timestamp": next_sunrise.astimezone(timezone.utc),
                "name": prefix,
                "value": "sunrise",
            }



import asyncio
import json
import logging
from datetime import datetime
from datetime import timezone
from typing import AsyncGenerator
from typing import List
from typing import Optional

from asyncio_mqtt import Client
from asyncio_mqtt import MqttError
from paho.mqtt.client import MQTTMessage
from senor_octopus.lib import merge_streams
from senor_octopus.types import Stream

_logger = logging.getLogger(__name__)

MessageStream = AsyncGenerator[MQTTMessage, None]

reconnect_interval = 3


async def mqtt(
    topics: List[str],
    host: str = "localhost",
    port: int = 1883,
    username: Optional[str] = None,
    password: Optional[str] = None,
    message_is_json: bool = False,
    prefix: str = "hub.mqtt",
) -> Stream:
    """
    Subscribe to messages on one or more MQTT topics.

    This source will subscribe to one or more MQTT topics (or topic
    wildcards), sending an event every time a message is received.

    Parameters
    ----------
    topics
        List of topics (or topic wildcards) to subscribe to
    host
        Host where the MQTT server is running
    port
        Port which the MQTT is listening to
    username
        Optional username to use when connecting to the MQTT server
    password
        Optional password to use when connecting to the MQTT server
    message_is_json
        The MQTT message is encoded as JSON and should be parsed
    prefix
        Prefix for events from this source

    Yields
    ------
    Event
        Events with data from MQTT messages
    """
    _logger.info("Connecting to MQTT server")

    while True:
        try:
            async with Client(
                host,
                port,
                username=username,
                password=password,
                client_id=prefix,  # XXX
                clean_session=False,
            ) as client:
                streams = [
                    read_from_topic(client, topic, prefix, message_is_json)
                    for topic in topics
                ]
                async for event in merge_streams(*streams):  # pragma: no cover
                    yield event
        except MqttError as error:
            _logger.warning(
                'Error "%s". Reconnecting in %s seconds.',
                error,
                reconnect_interval,
            )
            await asyncio.sleep(reconnect_interval)
        except asyncio.CancelledError:
            break


async def read_from_topic(
    client: Client,
    topic: str,
    prefix: str,
    message_is_json: bool,
) -> Stream:
    async with client.filtered_messages(topic) as messages:
        _logger.debug("Subscribing to topic: %s", topic)
        await client.subscribe(topic, qos=1)
        async for message in messages:  # pragma: no cover
            value = message.payload.decode()
            if message_is_json:
                try:
                    value = json.loads(value)
                except Exception:
                    pass

            yield {
                "timestamp": datetime.now(timezone.utc),
                "name": f"{prefix}.{message.topic}",
                "value": value,
            }



import logging
from datetime import datetime
from datetime import timezone
from typing import Generator

from senor_octopus.types import Event
from senor_octopus.types import Stream
from sqlalchemy import text
from sqlalchemy.engine import create_engine
from sqlalchemy.ext.asyncio import create_async_engine

_logger = logging.getLogger(__name__)


async def sqla(
    uri: str,
    sql: str,
    sync: bool = False,
    prefix: str = "hub.sqla",
) -> Stream:
    """
    Read data from database.

    The SQLAlchemy source periodically reads data from a database. It
    uses a query that MUST return at least two columns: `name` and `value`.
    Optionally, it an also return a column called `timestamp`, which will
    be used as the timestamp of the generated event. Otherwise, the current
    timestamp will be used.

    Parameters
    ----------
    uri
        SQLAlchemy URI (https://docs.sqlalchemy.org/en/14/core/engines.html)
    sql
        SQL query to run
    prefix
        Prefix for events from this source

    Yields
    ------
    Event
        Events with rows from database
    """
    _logger.info("Running SQL query")
    _logger.debug(sql)

    if sync:
        for event in read_sync(uri, sql, prefix):
            yield event
    else:
        async for event in read_async(uri, sql, prefix):  # pragma: no cover
            yield event


def read_sync(uri: str, sql: str, prefix: str) -> Generator[Event, None, None]:
    engine = create_engine(uri)

    with engine.connect() as conn:
        for row in conn.execute(text(sql)):
            _logger.debug(row)
            event = dict(row)
            yield {
                "timestamp": event.get("timestamp", datetime.now(timezone.utc)),
                "name": f"{prefix}.{event['name']}",
                "value": event["value"],
            }


async def read_async(uri: str, sql: str, prefix: str) -> Stream:
    engine = create_async_engine(uri)

    async with engine.connect() as conn:
        for row in await conn.execute(text(sql)):
            _logger.debug(row)
            event = dict(row)
            yield {
                "timestamp": event.get("timestamp", datetime.now(timezone.utc)),
                "name": f"{prefix}.{event['name']}",
                "value": event["value"],
            }



import logging
from datetime import datetime
from datetime import timezone

import httpx
from senor_octopus.lib import flatten
from senor_octopus.types import Stream

_logger = logging.getLogger(__name__)


async def weatherapi(
    access_token: str,
    location: str,
    prefix: str = "hub.weatherapi",
) -> Stream:
    """
    Fetch weather forecast data from weatherapi.com.

    This source will periodically retrieve weather forecast
    data from weatherapi.com for a given location.

    Parameters
    ----------
    access_token
        Access token from https://www.weatherapi.com/my/.
    location
        Location name or ZIP code
    prefix
        Prefix for events from this source

    Yields
    ------
    Event
        Events with weather forecast data
    """
    _logger.info("Fetching weather data")
    url = (
        f"http://api.weatherapi.com/v1/forecast.json?key={access_token}"
        f"&q={location}&days=2&aqi=no&alerts=no"
    )
    async with httpx.AsyncClient() as client:
        response = await client.get(url)
    payload = response.json()
    _logger.debug("Received %s", payload)

    for key, value in flatten(payload["current"]).items():
        yield {
            "timestamp": datetime.now(timezone.utc),
            "name": f"{prefix}.current.{key}",
            "value": value,
        }

    tomorrow = payload["forecast"]["forecastday"][1]
    for key, value in flatten(tomorrow["day"]).items():
        yield {
            "timestamp": datetime.now(timezone.utc),
            "name": f"{prefix}.forecast.forecastday.{key}",
            "value": value,
        }



from datetime import datetime
from typing import Any
from typing import AsyncGenerator

from typing_extensions import Protocol
from typing_extensions import TypedDict


class Event(TypedDict):
    timestamp: datetime
    name: str
    value: Any


Stream = AsyncGenerator[Event, None]


class SourceCallable(Protocol):
    def __call__(self, **kwargs: Any) -> Stream:
        ...  # pragma: no cover


class FilterCallable(Protocol):
    def __call__(self, stream: Stream, **kwargs: Any) -> Stream:
        ...  # pragma: no cover


class SinkCallable(Protocol):
    def __call__(self, stream: Stream, **kwargs: Any) -> None:
        ...  # pragma: no cover


class LoggerCallable(Protocol):
    def __call__(self, msg: str, *args: Any) -> None:
        ...  # pragma: no cover



import random
from datetime import datetime
from datetime import timezone

import pytest
from freezegun import freeze_time
from senor_octopus.filters.format import format
from senor_octopus.sources.rand import rand


@freeze_time("2021-01-01")
@pytest.mark.asyncio
async def test_format() -> None:
    random.seed(42)

    events = [event async for event in format(rand(2))]
    assert events == [
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.random",
            "value": 0.6394267984578837,
        },
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.random",
            "value": 0.025010755222666936,
        },
    ]

    events = [
        event
        async for event in format(
            rand(2),
            name="random number",
            value="{value:.2f}",
            eval_value=False,
        )
    ]
    assert events == [
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "random number",
            "value": "0.28",
        },
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "random number",
            "value": "0.22",
        },
    ]

    events = [
        event
        async for event in format(
            rand(2),
            value="{value:.2f}",
            eval_value=True,
        )
    ]
    assert events == [
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.random",
            "value": 0.74,
        },
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.random",
            "value": 0.68,
        },
    ]



"""
    Dummy conftest.py for senor_octopus.

    If you don't know what this is for, just leave it empty.
    Read more about conftest.py under:
    - https://docs.pytest.org/en/stable/fixture.html
    - https://docs.pytest.org/en/stable/writing_plugins.html
"""
import pytest
import yaml
from senor_octopus.graph import build_dag


config_content = """
random:
  plugin: source.random
  flow: -> check, normal
  schedule: "* * * * *"

check:
  plugin: filter.jsonpath
  flow: random -> high
  filter: $.events[?(@.value>0.5)]

normal:
  plugin: sink.log
  flow: random ->

high:
  plugin: sink.log
  flow: check ->
  throttle: 5 minutes
  level: warning
"""


@pytest.fixture
def mock_config():
    yield yaml.load(config_content)


@pytest.fixture
async def mock_dag(mock_config):
    dag = build_dag(mock_config)
    yield dag



import logging
from datetime import datetime
from datetime import timezone

import geohash
from aiohttp import ClientSession
from pywhistle import Client
from senor_octopus.types import Stream

_logger = logging.getLogger(__name__)


async def whistle(username: str, password: str, prefix: str = "hub.whistle") -> Stream:
    """
    Fetch device information and location for a Whistle pet tracker.

    This source will periodically fetch device information (battery level,
    last check-in, tracking status, battery status) and location for a
    given Whistle pet tracker.

    The location is emitted both as a pair of latitude, longitude, as well
    as a geohash.

    Parameters
    ----------
    username
        Username for https://app.whistle.com/
    password
        Password for https://app.whistle.com/
    prefix
        Prefix for events from this source

    Yields
    ------
    Event
        Events with device information and location
    """
    _logger.info("Fetching location from Whistle")
    async with ClientSession() as websession:
        client = Client(username, password, websession)
        await client.async_init()
        payload = await client.get_pets()
        _logger.debug("Received %s", payload)

        for pet in payload["pets"]:
            name = pet["name"]

            # device info
            keys = {
                "battery_level",
                "last_check_in",
                "tracking_status",
                "battery_status",
            }
            for key in keys:
                yield {
                    "timestamp": datetime.now(timezone.utc),
                    "name": f"{prefix}.{name}.{key}",
                    "value": pet["device"][key],
                }

            # last location
            location = pet["last_location"]
            yield {
                "timestamp": datetime.now(timezone.utc),
                "name": f"{prefix}.{name}.location",
                "value": (location["latitude"], location["longitude"]),
            }
            yield {
                "timestamp": datetime.now(timezone.utc),
                "name": f"{prefix}.{name}.geohash",
                "value": geohash.encode(location["latitude"], location["longitude"]),
            }



import asyncio
import logging
from typing import Any
from typing import cast
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple
from typing import Union

from asyncstdlib import itertools
from asyncstdlib.builtins import aiter
from crontab import CronTab
from durations import Duration
from pkg_resources import iter_entry_points
from senor_octopus.types import Event
from senor_octopus.types import FilterCallable
from senor_octopus.types import LoggerCallable
from senor_octopus.types import SinkCallable
from senor_octopus.types import SourceCallable
from senor_octopus.types import Stream


async def log_events(stream: Stream, flow: str, log: LoggerCallable) -> Stream:
    async for event in stream:
        log("%s: %s", flow, event)
        yield event


class Node:
    def __init__(self, node_name: str):
        self.name = node_name

        self.next: Set[Union["Filter", "Sink"]] = set()
        self._logger = logging.getLogger(node_name)
        self._event_logger = logging.getLogger("senor_octopus.events")

    @staticmethod
    def build(
        node_name: str,
        section: Dict[str, Any],
    ) -> Union["Source", "Filter", "Sink"]:
        # load plugin
        try:
            plugin_name = section.pop("plugin")
        except KeyError:
            raise Exception("Invalid config, missing `plugin` key")
        try:
            plugin = next(
                iter_entry_points("senor_octopus.plugins", plugin_name),
            ).load()
        except StopIteration:
            raise Exception(f"Invalid plugin name `{plugin_name}`")

        kwargs = section.copy()
        flow = kwargs.pop("flow").strip()
        if flow.startswith("->"):
            return Source(node_name, plugin, **kwargs)
        if flow.endswith("->"):
            return Sink(node_name, plugin, **kwargs)
        return Filter(node_name, plugin, **kwargs)


class Source(Node):
    def __init__(
        self,
        node_name: str,
        plugin: SourceCallable,
        schedule: Optional[str] = None,
        **extra_kwargs: Any,
    ):
        super().__init__(node_name)
        self.plugin = plugin
        self.schedule = CronTab(schedule) if schedule else None
        self.extra_kwargs = extra_kwargs

    async def run(self) -> None:
        self._logger.info("Running")
        stream = self.plugin(**self.extra_kwargs)
        async with itertools.tee(stream, n=len(self.next)) as streams:
            for node, stream in zip(self.next, streams):
                stream = log_events(
                    stream,
                    f"{self.name} -> {node.name}",
                    self._event_logger.debug,
                )
                await node.run(stream)


class Filter(Node):
    def __init__(self, node_name: str, plugin: FilterCallable, **extra_kwargs: Any):
        super().__init__(node_name)
        self.plugin = plugin
        self.extra_kwargs = extra_kwargs

    async def run(self, stream: Stream) -> None:
        self._logger.info("Running")
        stream = self.plugin(stream, **self.extra_kwargs)
        async with itertools.tee(stream, n=len(self.next)) as streams:
            for node, stream in zip(self.next, streams):
                stream = log_events(
                    stream,
                    f"{self.name} -> {node.name}",
                    self._event_logger.debug,
                )
                await node.run(stream)


class Sink(Node):
    def __init__(
        self,
        node_name: str,
        plugin: SinkCallable,
        throttle: Optional[str] = None,
        batch: Optional[str] = None,
        **extra_kwargs: Any,
    ):
        super().__init__(node_name)
        self.plugin = plugin
        self.throttle = Duration(throttle).to_seconds() if throttle else None
        self.batch = Duration(batch).to_seconds() if batch else None
        self.extra_kwargs = extra_kwargs

        self.last_run: Optional[float] = None
        self.queue: asyncio.Queue = asyncio.Queue()
        self.task = asyncio.create_task(self.worker())

    async def run(self, stream: Stream) -> None:
        loop = asyncio.get_running_loop()

        if (
            self.last_run is not None
            and self.throttle
            and loop.time() - self.last_run <= self.throttle
        ):
            self._logger.info(
                "Last run was %s, skipping this one due to throttle",
                self.last_run,
            )
            return

        self._logger.info("Running")
        stream = self.run_and_update_last_run(stream)

        # when in batch mode, send events to queue for worker to process
        if self.batch:
            self._logger.info("Sending events to queue")
            async for event in stream:  # pragma: no cover
                self.queue.put_nowait(event)
        else:
            self._logger.info("Processing events")
            await self.plugin(stream, **self.extra_kwargs)  # type: ignore

    async def run_and_update_last_run(self, stream: Stream) -> Stream:
        loop = asyncio.get_running_loop()

        at_least_one = False
        async for event in stream:  # pragma: no cover
            at_least_one = True
            yield event

        if at_least_one:
            self.last_run = loop.time()

    async def worker(self) -> None:
        if self.batch is None:
            return

        loop = asyncio.get_running_loop()
        canceled = False
        while not canceled:
            # consume queue into a new stream
            stream: List[Event] = []
            batch_start: Optional[float] = None
            while True:
                now = loop.time()

                timeout: Optional[float]
                if batch_start is None:
                    # batch hasn't started, we can wait forever on a new event
                    timeout = None
                else:
                    # wait on new event while the batch lasts
                    timeout = batch_start + self.batch - now

                try:
                    event = await asyncio.wait_for(self.queue.get(), timeout)
                except asyncio.TimeoutError:
                    break
                except RuntimeError:
                    return
                except asyncio.CancelledError:
                    self._logger.info("Cancelled, dumping currently batched events")
                    canceled = True
                    break

                # first event in a batch?
                if batch_start is None:
                    self._logger.info("Received event, starting a new batch")
                    batch_start = loop.time()

                stream.append(event)
                self.queue.task_done()

            self._logger.info("Processing batch")
            await self.plugin(aiter(stream), **self.extra_kwargs)  # type: ignore


def connected(config, source, target) -> bool:
    targets = config[source]["flow"].split("->")[1].strip()
    if targets != "*" and target not in {
        target.strip() for target in targets.split(",")
    }:
        return False

    sources = config[target]["flow"].split("->")[0].strip()
    if sources != "*" and source not in {
        source.strip() for source in sources.split(",")
    }:
        return False

    return True


def build_dag(config: Dict[str, Any]) -> Set[Source]:
    sections = set(config)
    for section in sections:
        if "flow" not in config[section]:
            raise Exception("Invalid config, missing `flow` key")

    sources = {
        name for name in sections if config[name]["flow"].strip().startswith("->")
    }
    sinks = {name for name in sections if config[name]["flow"].strip().endswith("->")}
    filters = sections - sources - sinks

    dag: Set[Source] = set()
    seen: Dict[str, Node] = {}
    queue: List[Tuple[Optional[str], str]] = [(None, name) for name in sources]
    while queue:
        source, target = queue.pop()
        if target in seen:
            node = seen[target]
        else:
            node = Node.build(target, config[target])
            seen[target] = node

        if source:
            node = cast(Union[Filter, Sink], node)
            seen[source].next.add(node)
        else:
            node = cast(Source, node)
            dag.add(node)

        source = target
        for target in filters | sinks:
            if connected(config, source, target):
                queue.append((source, target))

    return dag



from datetime import datetime
from datetime import timezone

import pytest
from aioresponses import aioresponses
from freezegun import freeze_time
from senor_octopus.sources.whistle import whistle

mock_payload = {
    "pets": [
        {
            "id": 123456,
            "gender": "f",
            "name": "Rex",
            "profile_photo_url_sizes": {
                "60x60": None,
                "100x100": None,
                "200x200": None,
                "750x750": None,
            },
            "realtime_channel": {"channel": "private-dog-123456", "service": "Pusher"},
            "subscription_status": "active",
            "partner_service_status": None,
            "device": {
                "model_id": "W04B",
                "serial_number": "W04-1234567",
                "last_check_in": "2021-03-25T15:52:19-07:00 America/Los_Angeles",
                "firmware_version": "0.47-200807 :: 0.47-200807 :: 0.47-200807",
                "battery_level": 96,
                "battery_status": "on",
                "pending_locate": False,
                "tracking_status": "not_tracking",
                "has_gps": True,
                "requires_subscription": True,
                "flashlight_status": {
                    "state": "off",
                    "pattern": None,
                    "brightness": None,
                },
                "partner_record": None,
            },
            "activity_summary": {
                "activity_start_date": "2020-08-22",
                "activity_enabled": True,
                "current_streak": 0,
                "current_minutes_active": 10,
                "current_minutes_rest": 827,
                "similar_dogs_minutes_active": 44.8529743653839,
                "similar_dogs_minutes_rest": 1083.1920639333,
                "suggested_activity_range_lower": 26.0,
                "suggested_activity_range_upper": 52.0,
                "current_activity_goal": {
                    "minutes": 36,
                    "started_at": "2020-08-22T07:00:00Z",
                    "time_zone": "America/Los_Angeles",
                },
                "upcoming_activity_goal": {
                    "minutes": 36,
                    "started_at": "2020-08-22T07:00:00Z",
                    "time_zone": "America/Los_Angeles",
                },
            },
            "last_location": {
                "latitude": 45.0,
                "longitude": -135.0,
                "timestamp": "2021-03-25T22:52:03Z",
                "uncertainty_meters": 0.0,
                "reason": "back_in_beacon",
                "place": {
                    "distance": 0,
                    "distance_units": "feet",
                    "id": 123456,
                    "status": "in_beacon_range",
                    "units": "feet",
                },
                "description": {
                    "address": "0 Fool's St",
                    "place": "Nowhere",
                    "postcode": "12345",
                    "region": "California",
                    "country": "United States",
                },
            },
            "profile": {
                "breed": {"id": 1870, "name": "Italian Greyhound Mix"},
                "date_of_birth": "2008-05-22",
                "age_in_months": 10,
                "age_in_years": 12,
                "time_zone_name": "America/Los_Angeles",
                "weight": 30.0,
                "weight_type": "pounds",
                "species": "dog",
                "overdue_task_occurrence_count": 4,
                "is_fixed": True,
                "body_condition_score": None,
                "pet_food": None,
            },
        },
    ],
}


@freeze_time("2021-01-01")
@pytest.mark.asyncio
async def test_whistle() -> None:
    with aioresponses() as mock_response:
        mock_response.post(
            "https://app.whistle.com/api/login",
            payload={"auth_token": "XXX"},
        )
        mock_response.get("https://app.whistle.com/api/pets", payload=mock_payload)
        events = [event async for event in whistle("username", "password")]

    print(sorted(events, key=lambda e: e["name"]))
    assert sorted(events, key=lambda e: e["name"]) == [
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.whistle.Rex.battery_level",
            "value": 96,
        },
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.whistle.Rex.battery_status",
            "value": "on",
        },
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.whistle.Rex.geohash",
            "value": "c00000000000",
        },
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.whistle.Rex.last_check_in",
            "value": "2021-03-25T15:52:19-07:00 America/Los_Angeles",
        },
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.whistle.Rex.location",
            "value": (45.0, -135.0),
        },
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.whistle.Rex.tracking_status",
            "value": "not_tracking",
        },
    ]



import asyncio
import logging
from typing import Awaitable
from typing import Dict
from typing import List
from typing import Set

from senor_octopus.graph import Source

_logger = logging.getLogger(__name__)


async def log_exceptions(run: Awaitable[None]) -> None:
    """
    Manually handle exceptions.

    https://bugs.python.org/issue39839
    """
    try:
        return await run
    except Exception:
        _logger.exception("Unhandled exception")
    return None


class Scheduler:
    def __init__(self, dag: Set[Source]):
        self.dag = dag
        self.tasks: List[asyncio.Task] = []
        self.cancelled = False

    async def run(self) -> None:
        if not self.dag:
            _logger.info("Nothing to run")
            return

        _logger.info("Starting scheduler")
        loop = asyncio.get_event_loop()

        event_nodes = {node for node in self.dag if not node.schedule}
        for node in event_nodes:
            _logger.debug("Starting %s", node.name)
            task = asyncio.create_task(log_exceptions(node.run()))
            self.tasks.append(task)

        schedule_nodes = {node for node in self.dag if node.schedule}
        schedules: Dict[str, float] = {}
        while not self.cancelled:
            for node in schedule_nodes:
                now = loop.time()
                delay = node.schedule.next(default_utc=False)
                when = now + delay

                if node.name in schedules and schedules[node.name] <= now:
                    _logger.info("Running %s", node.name)
                    task = asyncio.create_task(log_exceptions(node.run()))
                    self.tasks.append(task)
                    del schedules[node.name]

                if node.name not in schedules:
                    _logger.info("Scheduling %s to run in %d seconds", node.name, delay)
                    schedules[node.name] = when

            self.tasks = [task for task in self.tasks if not task.done()]

            sleep_time = min(schedules.values()) - now if schedules else 3600
            _logger.debug("Sleeping for %d seconds", sleep_time)
            await asyncio.sleep(sleep_time)

    def cancel(self) -> None:
        for task in self.tasks:
            task.cancel()
        self.cancelled = True



import logging
import textwrap

import aiopg
from psycopg2 import sql
from psycopg2.extras import Json
from senor_octopus.types import Stream

_logger = logging.getLogger(__name__)


async def postgresql(
    stream: Stream,
    user: str,
    password: str,
    host: str,
    port: int,
    dbname: str,
    table: str = "events",
) -> None:
    """
    Write events into a Postgres database.

    This sink can be used to write events into a Postgres database.
    It will create a table with 3 columns, `timestamp`, `name` and
    `event`, where to events will be stored.

    Parameters
    ----------
    stream
        The incoming stream of events
    user
        The username to use when connecting to the DB
    password
        The password to use when connecting to the DB
    host
        Host where the database is running
    post
        Port which the database is listening to
    dbname
        Name of the database to connect to
    table
        Name of the table where the events will be stored
    """
    dsn = f"dbname={dbname} user={user} password={password} host={host} port={port}"
    async with aiopg.create_pool(dsn) as pool:
        async with pool.acquire() as conn:
            async with conn.cursor() as cur:
                _logger.debug("Trying to create table `%s`", table)
                await cur.execute(
                    sql.SQL(
                        textwrap.dedent(
                            """
                            CREATE TABLE IF NOT EXISTS {table} (
                                "timestamp" TIMESTAMP,
                                "name" VARCHAR,
                                "value" JSON
                            );
                            """,
                        ),
                    ).format(table=sql.Identifier(table)),
                )
                _logger.debug("Trying to create index `%s_name_idx`", table)
                await cur.execute(
                    sql.SQL(
                        textwrap.dedent(
                            """
                            CREATE INDEX IF NOT EXISTS {index}
                            ON {table} USING HASH(name);
                            """,
                        ),
                    ).format(
                        table=sql.Identifier(table),
                        index=sql.Identifier(f"{table}_name_idx"),
                    ),
                )

                _logger.info("Inserting events into Postgres")
                async for event in stream:  # pragma: no cover
                    _logger.debug(event)
                    await cur.execute(
                        sql.SQL(
                            textwrap.dedent(
                                """
                                INSERT INTO {table} ("timestamp", "name", "value")
                                VALUES (%s, %s, %s);
                                """,
                            ),
                        ).format(table=sql.Identifier(table)),
                        (event["timestamp"], event["name"], Json(event["value"])),
                    )



import random

import pytest
from senor_octopus.sinks.slack import slack
from senor_octopus.sources.rand import rand


@pytest.mark.asyncio
async def test_slack(mocker) -> None:
    WebClient = mocker.patch("senor_octopus.sinks.slack.WebClient")
    client = WebClient.return_value

    random.seed(42)

    await slack(rand(1), "XXX", "C0123456789")

    client.chat_postMessage.assert_called_with(
        channel="C0123456789",
        text="0.6394267984578837",
    )



import asyncio
import random
from typing import cast

import aiotools
import pytest
import yaml
from freezegun import freeze_time
from senor_octopus.graph import build_dag
from senor_octopus.graph import connected
from senor_octopus.graph import Sink
from senor_octopus.graph import Source


def test_connected() -> None:
    config = {
        "a": {"flow": "-> *"},
        "b": {"flow": "-> c"},
        "c": {"flow": "* -> d, e"},
        "d": {"flow": "* ->"},
        "e": {"flow": "c, b ->"},
    }
    assert not connected(config, "a", "b")
    assert connected(config, "a", "c")
    assert connected(config, "a", "d")
    assert not connected(config, "a", "e")
    assert not connected(config, "b", "a")
    assert connected(config, "b", "c")
    assert not connected(config, "b", "d")
    assert not connected(config, "b", "e")
    assert not connected(config, "c", "a")
    assert not connected(config, "c", "b")
    assert connected(config, "c", "d")
    assert connected(config, "c", "e")

    # connections have a direction
    assert not connected(config, "d", "a")
    assert not connected(config, "d", "b")
    assert not connected(config, "d", "c")
    assert not connected(config, "d", "e")
    assert not connected(config, "e", "a")
    assert not connected(config, "e", "b")
    assert not connected(config, "e", "c")
    assert not connected(config, "e", "d")


@pytest.mark.asyncio
async def test_build_dag(mock_config) -> None:
    dag = build_dag(mock_config)
    assert len(dag) == 1

    source = dag.pop()
    assert isinstance(source, Source)
    assert len(source.next) == 2


@pytest.mark.asyncio
async def test_build_dag_many_to_one() -> None:
    config = yaml.load(
        """
one:
  plugin: source.random
  flow: -> three

two:
  plugin: source.random
  flow: -> three

three:
  plugin: sink.log
  flow: "* ->"
    """,
    )
    dag = build_dag(config)
    assert len(dag) == 2

    one, two = dag
    assert one.next == two.next


def test_build_dag_missing_plugin() -> None:
    config = yaml.load(
        """
a:
  flow: -> *
    """,
    )
    with pytest.raises(Exception) as excinfo:
        build_dag(config)
    assert str(excinfo.value) == "Invalid config, missing `plugin` key"


@pytest.mark.asyncio
async def test_build_dag_seen() -> None:
    config = yaml.load(
        """
a:
  flow: -> *
  plugin: source.random

b:
  flow: -> *
  plugin: source.random

c:
  flow: "* ->"
  plugin: sink.log
    """,
    )
    build_dag(config)


def test_build_dag_missing_flow() -> None:
    config = yaml.load(
        """
a:
  plugin: source.random
    """,
    )
    with pytest.raises(Exception) as excinfo:
        build_dag(config)
    assert str(excinfo.value) == "Invalid config, missing `flow` key"


def test_build_dag_invalid_plugin() -> None:
    config = yaml.load(
        """
a:
  plugin: source.invalid
  flow: -> *
    """,
    )
    with pytest.raises(Exception) as excinfo:
        build_dag(config)
    assert str(excinfo.value) == "Invalid plugin name `source.invalid`"


@freeze_time("2021-01-01")
@pytest.mark.asyncio
async def test_run_source(mocker, mock_config) -> None:
    mock_logger = mocker.patch("senor_octopus.sinks.log._logger")
    random.seed(42)

    dag = build_dag(mock_config)
    source = dag.pop()

    await source.run()
    assert len(mock_logger.log.mock_calls) == 14

    # test throttle, should get 10 entries from the non-throttled sink
    await source.run()
    assert len(mock_logger.log.mock_calls) == 24


@pytest.mark.asyncio
async def test_batch(mocker) -> None:
    mock_logger = mocker.patch("senor_octopus.sinks.log._logger")
    vclock = aiotools.VirtualClock()

    config = yaml.load(
        """
random:
  plugin: source.random
  flow: -> log
  schedule: "* * * * *"

log:
  plugin: sink.log
  flow: random ->
  batch: 2 minutes
    """,
    )

    with vclock.patch_loop():
        dag = build_dag(config)
        source = dag.pop()

        await source.run()
        assert len(mock_logger.log.mock_calls) == 0

        await asyncio.sleep(180)
        assert len(mock_logger.log.mock_calls) == 10


@pytest.mark.asyncio
async def test_batch_empty_source(mocker) -> None:
    mock_logger = mocker.patch("senor_octopus.sinks.log._logger")
    vclock = aiotools.VirtualClock()

    config = yaml.load(
        """
random:
  plugin: source.random
  flow: -> log
  schedule: "* * * * *"
  events: 0

log:
  plugin: sink.log
  flow: random ->
  batch: 2 minutes
    """,
    )

    with vclock.patch_loop():
        dag = build_dag(config)
        source = dag.pop()

        await source.run()
        assert len(mock_logger.log.mock_calls) == 0

        await asyncio.sleep(180)
        assert len(mock_logger.log.mock_calls) == 0


@pytest.mark.asyncio
async def test_throttle(mocker) -> None:
    _logger = mocker.patch("senor_octopus.sinks.log._logger")
    vclock = aiotools.VirtualClock()

    config = yaml.load(
        """
random:
  plugin: source.random
  flow: -> log
  schedule: "* * * * *"
  events: 1

log:
  plugin: sink.log
  flow: random ->
  throttle: 2 minutes
    """,
    )

    with vclock.patch_loop():
        dag = build_dag(config)
        source = dag.pop()
        sink = list(source.next)[0]

        sink = cast(Sink, sink)
        assert sink.last_run is None

        await source.run()
        assert len(_logger.log.mock_calls) == 1
        assert sink.last_run == 0.0

        await asyncio.sleep(60)
        await source.run()
        assert len(_logger.log.mock_calls) == 1
        assert sink.last_run == 0.0

        await asyncio.sleep(150)
        await source.run()
        assert len(_logger.log.mock_calls) == 2
        assert sink.last_run == 210.0


@pytest.mark.asyncio
async def test_throttle_without_events(mocker) -> None:
    _logger = mocker.patch("senor_octopus.sinks.log._logger")
    vclock = aiotools.VirtualClock()

    config = yaml.load(
        """
random:
  plugin: source.random
  flow: -> log
  schedule: "* * * * *"
  events: 0

log:
  plugin: sink.log
  flow: random ->
  throttle: 2 minutes
    """,
    )

    with vclock.patch_loop():
        dag = build_dag(config)
        source = dag.pop()
        sink = list(source.next)[0]

        sink = cast(Sink, sink)
        assert sink.last_run is None

        await source.run()
        assert len(_logger.log.mock_calls) == 0
        assert sink.last_run is None


@pytest.mark.asyncio
async def test_batch_cancel(mocker) -> None:
    mock_logger = mocker.MagicMock()
    mocker.patch("senor_octopus.sinks.log._logger", mock_logger)
    vclock = aiotools.VirtualClock()

    config = yaml.load(
        """
random:
  plugin: source.random
  flow: -> log
  schedule: "* * * * *"
  events: 0

log:
  plugin: sink.log
  flow: random ->
  batch: 2 minutes
    """,
    )

    with vclock.patch_loop():
        dag = build_dag(config)
        source = dag.pop()
        sink = list(source.next)[0]

        sink = cast(Sink, sink)
        sink.queue = mocker.MagicMock()
        sink.queue.get = mocker.AsyncMock(  # type: ignore
            side_effect=[0, 1, 2, asyncio.CancelledError("Cancelled")],
        )

        await source.run()
        assert len(mock_logger.log.mock_calls) == 0

        await asyncio.sleep(1800)
        assert len(mock_logger.log.mock_calls) == 3



import asyncio
from unittest import mock

import aiotools
import pytest
from asynctest import CoroutineMock
from senor_octopus.scheduler import Scheduler


@pytest.mark.asyncio
async def test_scheduler() -> None:
    mock_source1 = mock.MagicMock()
    mock_source1.schedule.next.return_value = 10
    mock_source1.run = CoroutineMock()
    mock_source1.run.side_effect = [None, None]
    mock_source2 = mock.MagicMock()
    mock_source2.schedule = None
    mock_source2.run = CoroutineMock()
    mock_dag = {mock_source1, mock_source2}
    vclock = aiotools.VirtualClock()

    with vclock.patch_loop():
        scheduler = Scheduler(mock_dag)  # type: ignore
        with pytest.raises(Exception) as excinfo:
            await scheduler.run()

    assert str(excinfo.value) == "coroutine raised StopIteration"
    assert len(mock_source1.run.mock_calls) == 3


@pytest.mark.asyncio
async def test_scheduler_short_long() -> None:
    mock_source1 = mock.MagicMock()
    mock_source1.schedule.next.return_value = 10
    mock_source1.run = CoroutineMock()
    mock_source1.run.side_effect = [None, None]
    mock_source2 = mock.MagicMock()
    mock_source2.schedule.next.return_value = 120
    mock_source2.run = CoroutineMock()
    mock_dag = {mock_source1, mock_source2}
    vclock = aiotools.VirtualClock()

    with vclock.patch_loop():
        scheduler = Scheduler(mock_dag)  # type: ignore
        with pytest.raises(Exception) as excinfo:
            await scheduler.run()

    assert str(excinfo.value) == "coroutine raised StopIteration"
    assert len(mock_source1.run.mock_calls) == 3


@pytest.mark.asyncio
async def test_scheduler_cancel() -> None:
    mock_source1 = mock.MagicMock()
    mock_source1.schedule.next.return_value = 10
    mock_source1.run = CoroutineMock()
    mock_dag = {mock_source1}
    vclock = aiotools.VirtualClock()

    async def cancel_scheduler(scheduler) -> None:
        await asyncio.sleep(30)
        scheduler.cancel()

    with vclock.patch_loop():
        scheduler = Scheduler(mock_dag)  # type: ignore
        await asyncio.gather(scheduler.run(), cancel_scheduler(scheduler))

    assert len(scheduler.tasks) == 1
    assert scheduler.cancelled
    assert len(mock_source1.run.mock_calls) == 2


@pytest.mark.asyncio
async def test_scheduler_no_jobs() -> None:
    vclock = aiotools.VirtualClock()
    with vclock.patch_loop():
        scheduler = Scheduler({})  # type: ignore
        await scheduler.run()


@pytest.mark.asyncio
async def test_scheduler_exceptions(mocker) -> None:
    mock_source1 = mock.MagicMock()
    mock_source1.schedule.next.return_value = 10
    mock_source1.run = CoroutineMock()
    mock_source1.run.side_effect = Exception("A wild error appeared!")
    mock_dag = {mock_source1}
    _logger = mocker.patch("senor_octopus.scheduler._logger")
    vclock = aiotools.VirtualClock()

    async def cancel_scheduler(scheduler) -> None:
        await asyncio.sleep(25)
        scheduler.cancel()

    with vclock.patch_loop():
        scheduler = Scheduler(mock_dag)  # type: ignore
        await asyncio.gather(scheduler.run(), cancel_scheduler(scheduler))

    assert len(_logger.exception.mock_calls) == 2



import random

import pytest
from senor_octopus.sinks.sms import sms
from senor_octopus.sources.rand import rand


@pytest.mark.asyncio
async def test_sms(mocker) -> None:
    Client = mocker.patch("senor_octopus.sinks.sms.Client")
    client = Client.return_value

    random.seed(42)

    await sms(rand(1), "alice", "XXX", "+15558675309", **{"from": "+18002738255"})

    client.messages.create.assert_called_with(
        body="0.6394267984578837",
        from_="+18002738255",
        to="+15558675309",
    )



import logging
from datetime import datetime
from datetime import timezone

import speedtest
from senor_octopus.types import Stream

_logger = logging.getLogger(__name__)


async def speed(prefix: str = "hub.speedtest") -> Stream:
    """
    Measure internet speed.

    This source will measure internet speed (download, upload,
    latency and more) by using the speedtest.net website.

    Parameters
    ----------
    prefix
        Prefix for events from this source

    Yields
    ------
    Event
        Events with internet speed data
    """
    _logger.info("Testing internet speed")
    s = speedtest.Speedtest()
    s.get_best_server()
    s.download()
    s.upload()
    _logger.debug("Received %s", s.results.dict())

    for key, value in s.results.dict().items():
        yield {
            "timestamp": datetime.now(timezone.utc),
            "name": f"{prefix}.{key}",
            "value": value,
        }



from datetime import datetime
from datetime import timezone

import pytest
from freezegun import freeze_time
from senor_octopus.sources.crypto import crypto

mock_payloads = [{"BTC": {"USD": 55816.61}}, {"AUDIO": {"USD": 4.051}}]


@freeze_time("2021-01-01")
@pytest.mark.asyncio
async def test_crypto(mocker) -> None:
    mock = mocker.patch("senor_octopus.sources.crypto.cryptocompare")

    mock.get_price.side_effect = mock_payloads
    coins = ["BTC", "AUDIO"]
    events = [event async for event in crypto(coins)]
    assert events == [
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.crypto.BTC.USD",
            "value": 55816.61,
        },
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.crypto.AUDIO.USD",
            "value": 4.051,
        },
    ]



from datetime import datetime
from datetime import timezone

import pytest
from freezegun import freeze_time
from senor_octopus.sources.sqla import sqla


results = [
    {
        "timestamp": datetime(2020, 12, 31, 0, 0, tzinfo=timezone.utc),
        "name": "foo",
        "value": "bar",
    },
    {"name": "foo", "value": "baz"},
]


@freeze_time("2021-01-01")
@pytest.mark.asyncio
async def test_sqla(mocker) -> None:
    mock = mocker.patch("senor_octopus.sources.sqla.create_async_engine")
    mock_conn = (
        mock.return_value.connect.return_value.__aenter__.return_value
    ) = mocker.MagicMock()
    mock_conn.execute = mocker.AsyncMock()
    mock_conn.execute.return_value = results

    events = [event async for event in sqla("uri", "sql")]
    assert events == [
        {
            "timestamp": datetime(2020, 12, 31, 0, 0, tzinfo=timezone.utc),
            "name": "hub.sqla.foo",
            "value": "bar",
        },
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.sqla.foo",
            "value": "baz",
        },
    ]


@freeze_time("2021-01-01")
@pytest.mark.asyncio
async def test_sqla_sync(mocker) -> None:
    mock = mocker.patch("senor_octopus.sources.sqla.create_engine")
    mock_conn = (
        mock.return_value.connect.return_value.__enter__.return_value
    ) = mocker.MagicMock()
    mock_conn.execute = mocker.Mock()
    mock_conn.execute.return_value = results

    events = [event async for event in sqla("uri", "sql", sync=True)]
    assert events == [
        {
            "timestamp": datetime(2020, 12, 31, 0, 0, tzinfo=timezone.utc),
            "name": "hub.sqla.foo",
            "value": "bar",
        },
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.sqla.foo",
            "value": "baz",
        },
    ]



import random

import pytest
from senor_octopus.sinks.pushover import pushover
from senor_octopus.sources.rand import rand


@pytest.mark.asyncio
async def test_pushover(httpx_mock) -> None:
    httpx_mock.add_response()
    random.seed(42)

    await pushover(rand(2), "XXX", "alice")
    requests = httpx_mock.get_requests()
    assert len(requests) == 2
    assert (
        requests[0].read()
        == b"token=XXX&user=alice&message=hub.random%3A+0.6394267984578837"
    )


@pytest.mark.asyncio
async def test_pushover_empty_stream(httpx_mock) -> None:
    random.seed(42)

    await pushover(rand(0), "XXX", "alice")
    assert not httpx_mock.get_request()



from datetime import datetime
from datetime import timezone

import pytest
from freezegun import freeze_time
from senor_octopus.sources.static import static


@freeze_time("2021-01-01T12:00:00-07:00")
@pytest.mark.asyncio
async def test_sun() -> None:
    events = [event async for event in static("name", "value")]

    print(events)
    assert events == [
        {
            "timestamp": datetime(2021, 1, 1, 19, tzinfo=timezone.utc),
            "name": "name",
            "value": "value",
        },
    ]



import random
from datetime import datetime
from datetime import timezone

from senor_octopus.types import Stream


async def rand(events: int = 10, prefix: str = "hub.random") -> Stream:
    """
    Generate random numbers between 0 and 1.

    This source will generate random numbers between 0 and 1
    when schedule. It's useful for testing.

    Parameters
    ----------
    events
        Number of events to generate every time it runs
    prefix
        Prefix for events from this source

    Yields
    ------
    Event
         Events with random numbers
    """
    for _ in range(events):
        yield {
            "timestamp": datetime.now(timezone.utc),
            "name": prefix,
            "value": random.random(),
        }



import random

import pytest
from senor_octopus.sinks.mqtt import mqtt
from senor_octopus.sources.rand import rand


@pytest.mark.asyncio
async def test_mqtt(mocker) -> None:
    mock = mocker.patch("senor_octopus.sinks.mqtt.Client")
    mock__publish = (
        mock.return_value.__aenter__.return_value.publish
    ) = mocker.AsyncMock()
    mock.return_value.__aexit__ = mocker.AsyncMock()
    random.seed(42)

    await mqtt(rand(2), "topic")

    mock__publish.assert_has_calls(
        [
            mocker.call("topic", 0.6394267984578837, qos=1),
            mocker.call("topic", 0.025010755222666936, qos=1),
        ],
    )



from datetime import datetime
from datetime import timezone

import aiotools
import pytest
from freezegun import freeze_time
from senor_octopus.sources.sun import sun


@freeze_time("2021-01-01T12:00:00-07:00")
@pytest.mark.asyncio
async def test_sun() -> None:
    vclock = aiotools.VirtualClock()

    latitude = 38.3
    longitude = -123.0

    with vclock.patch_loop():
        events = sun(latitude, longitude)
        event = await events.__anext__()

    assert event == {
        "timestamp": datetime(2021, 1, 2, 1, 3, tzinfo=timezone.utc),
        "name": "hub.sun",
        "value": "sunset",
    }


@freeze_time("2021-01-01T12:00:00-07:00")
@pytest.mark.asyncio
async def test_sun_east() -> None:
    vclock = aiotools.VirtualClock()

    latitude = 51.21
    longitude = 21.01

    with vclock.patch_loop():
        events = sun(latitude, longitude)
        event = await events.__anext__()

    assert event == {
        "timestamp": datetime(2021, 1, 2, 6, 41, tzinfo=timezone.utc),
        "name": "hub.sun",
        "value": "sunrise",
    }


@freeze_time("2021-01-01T01:00:00-07:00")
@pytest.mark.asyncio
async def test_sun_early() -> None:
    vclock = aiotools.VirtualClock()

    latitude = 38.3
    longitude = -123.0

    with vclock.patch_loop():
        events = sun(latitude, longitude)
        event = await events.__anext__()

    assert event == {
        "timestamp": datetime(2021, 1, 1, 15, 29, tzinfo=timezone.utc),
        "name": "hub.sun",
        "value": "sunrise",
    }


@freeze_time("2021-01-01T23:00:00-07:00")
@pytest.mark.asyncio
async def test_sun_late() -> None:
    vclock = aiotools.VirtualClock()

    latitude = 38.3
    longitude = -123.0

    with vclock.patch_loop():
        events = sun(latitude, longitude)
        event = await events.__anext__()

    assert event == {
        "timestamp": datetime(2021, 1, 2, 15, 29, tzinfo=timezone.utc),
        "name": "hub.sun",
        "value": "sunrise",
    }



import logging
from typing import Optional

from asyncio_mqtt import Client
from senor_octopus.types import Stream

_logger = logging.getLogger(__name__)


async def mqtt(
    stream: Stream,
    topic: str,
    host: str = "localhost",
    port: int = 1883,
    username: Optional[str] = None,
    password: Optional[str] = None,
    qos: int = 1,
) -> None:
    """
    Send events as messages to an MQTT topic.

    This sink can be used to send events to an MQTT topic.
    The value of the event is sent as the message; its name
    is ignored.

    Parameters
    ----------
    stream
        The incoming stream of events
    topic
        The MQTT topic where messages are sent to
    host
        Host where the MQTT server is running
    port
        Port which the MQTT is listening to
    username
        Optional username to use when connecting to the MQTT server
    password
        Optional password to use when connecting to the MQTT server
    qos
        Quality of Service (QoS) level
    """
    async with Client(host, port, username=username, password=password) as client:
        async for event in stream:  # pragma: no cover
            await client.publish(topic, event["value"], qos=qos)



import random
from datetime import datetime
from datetime import timezone

import pytest
from freezegun import freeze_time
from senor_octopus.filters.jinja import jinja
from senor_octopus.sources.rand import rand


@freeze_time("2021-01-01")
@pytest.mark.asyncio
async def test_jinja() -> None:
    random.seed(42)

    template = (
        "{% if event['value'] < 0.5 %}{{ '{:.2f}'.format(event['value']) }}{% endif %}"
    )
    events = [event async for event in jinja(rand(2), template)]
    assert events == [
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.random",
            "value": "0.03",
        },
    ]



import asyncio
from dataclasses import dataclass
from datetime import datetime
from datetime import timezone

import aiotools
import pytest
from asyncio_mqtt import MqttError
from freezegun import freeze_time
from senor_octopus.sources.mqtt import mqtt


@dataclass
class FakeMessage:
    topic: str
    payload: bytes


@freeze_time("2021-01-01")
@pytest.mark.asyncio
async def test_mqtt(mocker) -> None:
    vclock = aiotools.VirtualClock()

    async def gen(events, exception):
        for i in range(events):
            yield FakeMessage("topic/1", str(i).encode("utf-8"))

        raise exception

    client1 = mocker.AsyncMock()
    client1.__aenter__.return_value = mocker.MagicMock()
    client1.__aenter__.return_value.subscribe = mocker.AsyncMock()
    filtered_messages = client1.__aenter__.return_value.filtered_messages
    filtered_messages.return_value.__aenter__ = mocker.AsyncMock()
    filtered_messages.return_value.__aenter__.return_value = gen(
        3,
        MqttError("Disconnect"),
    )

    client2 = mocker.AsyncMock()
    client2.__aenter__.return_value = mocker.MagicMock()
    client2.__aenter__.return_value.subscribe = mocker.AsyncMock()
    filtered_messages = client2.__aenter__.return_value.filtered_messages
    filtered_messages.return_value.__aenter__ = mocker.AsyncMock()
    filtered_messages.return_value.__aenter__.return_value = gen(
        0,
        MqttError("Disconnect"),
    )

    client3 = mocker.AsyncMock()
    client3.__aenter__.return_value = mocker.MagicMock()
    client3.__aenter__.return_value.subscribe = mocker.AsyncMock()
    filtered_messages = client3.__aenter__.return_value.filtered_messages
    filtered_messages.return_value.__aenter__ = mocker.AsyncMock()
    filtered_messages.return_value.__aenter__.return_value = gen(
        3,
        asyncio.CancelledError("Canceled"),
    )

    Client = mocker.patch("senor_octopus.sources.mqtt.Client")
    Client.side_effect = [client1, client2, client3]

    topics = ["topic/#"]
    with vclock.patch_loop():
        events = [event async for event in mqtt(topics)]
    assert events == [
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.mqtt.topic/1",
            "value": "0",
        },
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.mqtt.topic/1",
            "value": "1",
        },
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.mqtt.topic/1",
            "value": "2",
        },
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.mqtt.topic/1",
            "value": "0",
        },
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.mqtt.topic/1",
            "value": "1",
        },
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.mqtt.topic/1",
            "value": "2",
        },
    ]


@freeze_time("2021-01-01")
@pytest.mark.asyncio
async def test_mqtt_multiple_topics(mocker) -> None:
    vclock = aiotools.VirtualClock()

    mock = mocker.patch("senor_octopus.sources.mqtt.Client")
    mock.return_value.__aenter__.return_value = mocker.MagicMock()
    mock.return_value.__aenter__.return_value.subscribe = mocker.AsyncMock()

    async def gen(topic):
        sleep = 2 if topic == "topic/1" else 5
        count = 3 if topic == "topic/1" else 2

        for i in range(count):
            await asyncio.sleep(sleep)
            yield FakeMessage(topic, str(i).encode("utf-8"))

        if topic == "other/1":
            raise asyncio.CancelledError("Canceled")

    def get_stream(topic):
        stream = mocker.MagicMock()
        stream.__aenter__ = mocker.AsyncMock()
        stream.__aenter__.return_value = gen(topic)
        return stream

    mock.return_value.__aenter__.return_value.filtered_messages.side_effect = get_stream
    topics = ["topic/1", "other/1"]
    with vclock.patch_loop():
        events = [event async for event in mqtt(topics, message_is_json=True)]
    assert events == [
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.mqtt.topic/1",
            "value": 0,
        },
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.mqtt.topic/1",
            "value": 1,
        },
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.mqtt.other/1",
            "value": 0,
        },
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.mqtt.topic/1",
            "value": 2,
        },
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.mqtt.other/1",
            "value": 1,
        },
    ]



import ast
import logging
from typing import Optional

from senor_octopus.types import Stream

_logger = logging.getLogger(__name__)


async def format(
    stream: Stream,
    name: Optional[str] = None,
    value: Optional[str] = None,
    eval_value: bool = False,
) -> Stream:
    """
    Format an event stream based using Python string formatting.

    This filter will replace the name and/or value of an event
    by formatting a string based on the original event.

    Parameters
    ----------
    stream
        The incoming stream of events
    name
        A string formatted with the original event that replaces the name
    value
        A string formatted with the original event that replaces the value
    eval_value
        Wether the new value should be evaluated

    Yields
    ------
    Event
        Events formatted according to the configuration
    """
    _logger.debug("Formatting events")
    async for event in stream:  # pragma: no cover
        event = event.copy()
        if name:
            event["name"] = name.format(**event)
        if value:
            event["value"] = value.format(**event)
        if eval_value:
            event["value"] = ast.literal_eval(event["value"])

        yield event



import logging

from senor_octopus.types import Stream
from twilio.rest import Client

_logger = logging.getLogger(__name__)


async def sms(
    stream: Stream, account_sid: str, auth_token: str, to: str, **kwargs: str
) -> None:
    """
    Send SMS via Twilio.

    Parameters
    ----------
    stream
        The incoming stream of events
    account_sid
        The account SID (https://www.twilio.com/console)
    auth_token
        The auth token (https://www.twilio.com/console)
    from
        Twilio phone number
    to
        Recipient phone number
    """
    from_ = kwargs["from"]
    client = Client(account_sid, auth_token)

    async for event in stream:  # pragma: no cover
        _logger.debug(event)
        _logger.info("Sending SMS")
        client.messages.create(body=str(event["value"]).strip(), from_=from_, to=to)



import random
from datetime import datetime
from datetime import timezone
from unittest import mock

import pytest
from asynctest import CoroutineMock
from freezegun import freeze_time
from psycopg2 import sql
from senor_octopus.sinks.db.postgresql import postgresql
from senor_octopus.sources.rand import rand


@freeze_time("2021-01-01")
@pytest.mark.asyncio
async def test_postgresql(mocker) -> None:
    mock_aiopg = CoroutineMock()
    pool = mock_aiopg.create_pool.return_value.__aenter__.return_value
    conn = pool.acquire.return_value.__aenter__.return_value
    cursor = conn.cursor.return_value.__aenter__.return_value
    cursor.execute = CoroutineMock()
    mocker.patch("senor_octopus.sinks.db.postgresql.aiopg", mock_aiopg)
    random.seed(42)

    await postgresql(rand(1), "user", "password", "host", 5432, "dbname")

    cursor.execute.assert_has_calls(
        [
            mock.call(
                sql.Composed(
                    [
                        sql.SQL("\nCREATE TABLE IF NOT EXISTS "),
                        sql.Identifier("events"),
                        sql.SQL(
                            ' (\n    "timestamp" TIMESTAMP,\n    '
                            '"name" VARCHAR,\n    "value" JSON\n);\n',
                        ),
                    ],
                ),
            ),
            mock.call(
                sql.Composed(
                    [
                        sql.SQL("\nCREATE INDEX IF NOT EXISTS "),
                        sql.Identifier("events_name_idx"),
                        sql.SQL("\nON "),
                        sql.Identifier("events"),
                        sql.SQL(" USING HASH(name);\n"),
                    ],
                ),
            ),
            mock.call(
                sql.Composed(
                    [
                        sql.SQL("\nINSERT INTO "),
                        sql.Identifier("events"),
                        sql.SQL(
                            ' ("timestamp", "name", "value")\nVALUES (%s, %s, %s);\n',
                        ),
                    ],
                ),
                (
                    datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
                    "hub.random",
                    mock.ANY,  # Json(0.6394267984578837)
                ),
            ),
        ],
    )


@pytest.mark.asyncio
async def test_postgresql_empty_stream(mocker) -> None:
    mock_aiopg = CoroutineMock()
    pool = mock_aiopg.create_pool.return_value.__aenter__.return_value
    conn = pool.acquire.return_value.__aenter__.return_value
    cursor = conn.cursor.return_value.__aenter__.return_value
    cursor.execute = CoroutineMock()
    mocker.patch("senor_octopus.sinks.db.postgresql.aiopg", mock_aiopg)
    random.seed(42)

    await postgresql(rand(0), "user", "password", "host", 5432, "dbname")
    assert len(cursor.execute.mock_calls) == 2



from datetime import datetime
from datetime import timezone

import pytest
from senor_octopus.sources.awair import awair

mock_payload = {
    "data": [
        {
            "timestamp": "2021-03-23T22:26:51.000Z",
            "score": 95.0,
            "sensors": [
                {"comp": "temp", "value": 25.0},
                {"comp": "voc", "value": 158.0},
                {"comp": "pm25", "value": 3.0},
                {"comp": "co2", "value": 640.0},
                {"comp": "humid", "value": 42.45000076293945},
            ],
            "indices": [
                {"comp": "pm25", "value": 0.0},
                {"comp": "voc", "value": 0.0},
                {"comp": "humid", "value": 0.0},
                {"comp": "co2", "value": 1.0},
                {"comp": "temp", "value": 0.0},
            ],
        },
    ],
}


@pytest.mark.asyncio
async def test_awair(httpx_mock) -> None:
    httpx_mock.add_response(json=mock_payload)

    events = [event async for event in awair("XXX", 12345)]
    assert events == [
        {
            "timestamp": datetime(2021, 3, 23, 22, 26, 51, tzinfo=timezone.utc),
            "name": "hub.awair.score",
            "value": 95.0,
        },
        {
            "timestamp": datetime(2021, 3, 23, 22, 26, 51, tzinfo=timezone.utc),
            "name": "hub.awair.temp",
            "value": 25.0,
        },
        {
            "timestamp": datetime(2021, 3, 23, 22, 26, 51, tzinfo=timezone.utc),
            "name": "hub.awair.voc",
            "value": 158.0,
        },
        {
            "timestamp": datetime(2021, 3, 23, 22, 26, 51, tzinfo=timezone.utc),
            "name": "hub.awair.pm25",
            "value": 3.0,
        },
        {
            "timestamp": datetime(2021, 3, 23, 22, 26, 51, tzinfo=timezone.utc),
            "name": "hub.awair.co2",
            "value": 640.0,
        },
        {
            "timestamp": datetime(2021, 3, 23, 22, 26, 51, tzinfo=timezone.utc),
            "name": "hub.awair.humid",
            "value": 42.45000076293945,
        },
    ]



# flake8: noqa
import asyncio
import random

import aiotools
import pytest
import yaml
from senor_octopus.graph import build_dag
from senor_octopus.lib import flatten
from senor_octopus.lib import merge_streams
from senor_octopus.lib import render_dag
from senor_octopus.sources.rand import rand


def test_flatten() -> None:
    assert flatten({"a": 1, "b": 2.0, "c": "foo", "d": {"e": "bar"}}) == {
        "a": 1,
        "b": 2.0,
        "c": "foo",
        "d.e": "bar",
    }


def test_render_dag(mock_dag) -> None:
    assert (
        render_dag(mock_dag, use_color=False).strip()
        == r"""
*   random
|\  
* | check
| * normal
* high
        """.strip()
    )


@pytest.mark.asyncio
async def test_render_dag_many_to_many() -> None:
    config = yaml.load(
        """
one:
  plugin: source.random
  flow: -> two, three

two:
  plugin: filter.jsonpath
  flow: one -> four
  filter: $.events[?(@.value>0.5)]

three:
  plugin: filter.jsonpath
  flow: one -> four
  filter: $.events[?(@.value<=0.5)]

four:
  plugin: sink.log
  flow: two, three ->
        """,
    )

    dag = build_dag(config)
    assert (
        render_dag(dag, use_color=False).strip()
        == r"""
*   one
|\  
* | three
| * two
|/  
* four
        """.strip()
    )


@pytest.mark.asyncio
async def test_merge_streams() -> None:
    vclock = aiotools.VirtualClock()

    random.seed(42)

    async def gen(name, count, sleep):
        for i in range(count):
            await asyncio.sleep(sleep)
            yield (name, random.random())

    stream = gen("stream1", 3, 2)
    with vclock.patch_loop():
        events = [event async for event in stream]
    assert events == [
        ("stream1", 0.6394267984578837),
        ("stream1", 0.025010755222666936),
        ("stream1", 0.27502931836911926),
    ]

    stream = gen("stream2", 2, 5)
    with vclock.patch_loop():
        events = [event async for event in stream]
    assert events == [("stream2", 0.22321073814882275), ("stream2", 0.7364712141640124)]

    random.seed(42)
    stream = merge_streams(gen("stream1", 3, 2), gen("stream2", 2, 5))
    with vclock.patch_loop():
        events = [event async for event in stream]
    assert events == [
        ("stream1", 0.6394267984578837),
        ("stream1", 0.025010755222666936),
        ("stream2", 0.27502931836911926),
        ("stream1", 0.22321073814882275),
        ("stream2", 0.7364712141640124),
    ]



import logging

import httpx
from senor_octopus.types import Stream

_logger = logging.getLogger(__name__)


async def pushover(stream: Stream, app_token: str, user_token: str) -> None:
    """
    Send events to the Pushover mobile app.

    This sink can be used to send events to a smartphone, using
    the Pushover (pushover.net) application.

    Parameters
    ----------
    stream
        The incoming stream of events
    app_token
        The application token (https://pushover.net/api)
    user_token
        The user token (https://pushover.net/api)
    """
    url = "https://api.pushover.net/1/messages.json"
    async for event in stream:  # pragma: no cover
        _logger.debug(event)
        data = {
            "token": app_token,
            "user": user_token,
            "message": "{name}: {value}".format(**event),
        }
        _logger.info("Posting message to Pushover")
        async with httpx.AsyncClient() as client:
            await client.post(url, data=data)



import logging
from datetime import datetime
from datetime import timezone
from typing import List

import cryptocompare
from senor_octopus.types import Stream

_logger = logging.getLogger(__name__)


async def crypto(
    coins: List[str],
    currency: str = "USD",
    prefix: str = "hub.crypto",
) -> Stream:
    """
    Fetch price of cryptocurrencies from cryptocompare.com.

    This source will periodically retrieve the price of
    cryptocurrencies from the cryptocompare.com website.

    Parameters
    ----------
    coins
        List of cryptocurrency symbols
    currency
        Currency used to display price
    prefix
        Prefix for events from this source

    Yields
    ------
    Event
        Events with price of cryptocurrencies
    """
    _logger.info("Fetching crypto data")

    for coin in coins:
        info = cryptocompare.get_price(coin, currency=currency, full=False)
        value = info[coin][currency]
        _logger.debug("%s: %s %s", coin, currency, value)
        yield {
            "timestamp": datetime.now(timezone.utc),
            "name": f"{prefix}.{coin}.{currency}",
            "value": value,
        }



from datetime import datetime
from datetime import timezone

from senor_octopus.types import Stream


async def static(name: str, value: str) -> Stream:
    """
    Generate static events.

    Parameters
    ----------
    name
        Name of the event
    value
        Value of the event

    Yields
    ------
    Event
        Static event
    """
    yield {
        "timestamp": datetime.now(timezone.utc),
        "name": name,
        "value": value,
    }



from datetime import datetime
from datetime import timezone

import pytest
from freezegun import freeze_time
from senor_octopus.sources.stock import stock


@freeze_time("2021-01-01")
@pytest.mark.asyncio
async def test_stock(mocker) -> None:
    mock = mocker.patch("senor_octopus.sources.stock.stockquotes")
    type(mock.Stock.return_value).current_price = mocker.PropertyMock(
        side_effect=[283.02, 64.51, 283.02, 64.51],
    )
    type(mock.Stock.return_value).increase_percent = mocker.PropertyMock(
        side_effect=[1.54, 1.48, 1.54, 1.48],
    )
    symbols = ["FB", "LYFT"]
    events = [event async for event in stock(symbols)]
    events = sorted(events, key=lambda e: e["name"])
    assert events == [
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.stock.FB.current_price",
            "value": 283.02,
        },
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.stock.FB.increase_percent",
            "value": 1.54,
        },
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.stock.LYFT.current_price",
            "value": 64.51,
        },
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.stock.LYFT.increase_percent",
            "value": 1.48,
        },
    ]



import logging

from jsonpath import JSONPath
from senor_octopus.types import Stream

_logger = logging.getLogger(__name__)


async def jsonpath(stream: Stream, filter: str) -> Stream:
    """
    Filter event stream based on a JSON path.

    This filter can be used to filter events based on their
    name and value, allowing events to be dynamically routed.

    Parameters
    ----------
    stream
        The incoming stream of events
    filter
        A JSON Path expression used to filter events

    Yields
    ------
    Event
        Events filtered by the JSON Path expression
    """
    _logger.debug("Filtering events")
    parser = JSONPath(filter)
    async for event in stream:  # pragma: no cover
        if parser.parse({"events": [event]}):
            yield event



import logging
from functools import lru_cache

from senor_octopus.types import Stream
from tuyapy import TuyaApi
from typing_extensions import Literal

_logger = logging.getLogger(__name__)


@lru_cache(maxsize=None)
def authenticate(email: str, password: str, country: str, application: str) -> TuyaApi:
    _logger.debug("Authenticating")
    api = TuyaApi()
    api.init(email, password, country, application)
    return api


async def tuya(
    stream: Stream,
    device: str,
    email: str,
    password: str,
    country: str = "1",
    application: str = Literal["smart_life", "tuya"],
) -> None:
    """
    Send commands to a Tuya/Smart Life device.

    Currently this plugin supports sending on and off events, but
    it can be easily modified to support changing the color of a
    lightbulb.

    Parameters
    ----------
    stream
        The incoming stream of events
    device
        The name of the device to be controlled
    email
        The email of the account
    password
        The password of the account
    country
        Country telephone code
    application
        The application code, either "tuya" or "smart_life"
    """
    api = authenticate(email, password, country, application)
    devices = {d.name(): d for d in api.get_all_devices()}
    if device not in devices:
        valid = ", ".join(f'"{name}"' for name in devices)
        _logger.error('Device "%s" not found. Available devices: %s', device, valid)
        return

    async for event in stream:  # pragma: no cover
        _logger.debug(event)
        if event["value"].lower() == "on":
            devices[device].turn_on()
        elif event["value"].lower() == "off":
            devices[device].turn_off()
        else:
            _logger.warning("Unknown value: %s", event["value"])



import logging
from datetime import datetime
from datetime import timezone
from typing import List

import stockquotes
from senor_octopus.types import Stream

_logger = logging.getLogger(__name__)


async def stock(symbols: List[str], prefix: str = "hub.stock") -> Stream:
    """
    Fetch stock price form Yahoo! Finance.

    This source will periodically retrieve the price of
    stocks from Yahoo! Finance.

    Parameters
    ----------
    symbols
        List of stock symbols
    prefix
        Prefix for events from this source

    Yields
    ------
    Event
        Events with price of stocks
    """
    _logger.info("Fetching stock data")

    for symbol in symbols:
        ticker = stockquotes.Stock(symbol)
        for attribute in {"current_price", "increase_percent"}:
            value = getattr(ticker, attribute)
            _logger.debug(
                "%s (%s) %s: %s",
                ticker.symbol,
                ticker.name,
                attribute,
                value,
            )
            yield {
                "timestamp": datetime.now(timezone.utc),
                "name": f"{prefix}.{symbol}.{attribute}",
                "value": value,
            }



import logging

from senor_octopus.types import Stream

_logger = logging.getLogger(__name__)


async def log(stream: Stream, level: str = "INFO") -> None:
    """
    Send events to a logger.

    This sink can be used to send events to a logger, using
    a configurable level (DEBUG, INFO, WARNING, ERROR, etc.).

    Parameters
    ----------
    stream
        The incoming stream of events
    level
        The logging level to be used
    """
    async for event in stream:
        _logger.log(getattr(logging, level.upper()), event)



import logging

import dateutil.parser
import httpx
from senor_octopus.types import Stream

_logger = logging.getLogger(__name__)


async def awair(
    access_token: str,
    device_id: int,
    device_type: str = "awair-element",
    prefix: str = "hub.awair",
) -> Stream:
    """
    Fetch air quality data from Awair Element monitor.

    This source will periodically retrieve air quality data from
    an Awair Element monitor.

    Parameters
    ----------
    access_token
        Awair API access token (https://developer.getawair.com/console/access-token)
    device_id
        Device ID
    device_type
        Device type
    prefix
        Prefix for events from this source

    Yields
    ------
    Event
        Events with data from sensors
    """
    _logger.info("Fetching air quality data")
    url = (
        "https://developer-apis.awair.is/v1/users/self/devices/"
        f"{device_type}/{device_id}/air-data/latest?fahrenheit=false"
    )
    headers = {"Authorization": f"Bearer {access_token}"}
    async with httpx.AsyncClient() as client:
        response = await client.get(url, headers=headers)
    payload = response.json()
    _logger.debug("Received %s", payload)

    for row in payload["data"]:
        timestamp = dateutil.parser.parse(row["timestamp"])
        yield {
            "timestamp": timestamp,
            "name": f"{prefix}.score",
            "value": row["score"],
        }

        for sensor in row["sensors"]:
            yield {
                "timestamp": timestamp,
                "name": f"{prefix}.{sensor['comp']}",
                "value": sensor["value"],
            }



from datetime import datetime
from datetime import timezone
from unittest import mock

import pytest
from freezegun import freeze_time
from senor_octopus.sources.speed import speed

mock_payload = {
    "download": 20932433.00558606,
    "upload": 6068322.252775613,
    "ping": 50.377,
    "server": {
        "url": "http://speedtest1.st-tel.net:8080/speedtest/upload.php",
        "lat": "39.3958",
        "lon": "-101.0519",
        "name": "Colby, KS",
        "country": "United States",
        "cc": "US",
        "sponsor": "S&T Communications",
        "id": "3434",
        "host": "speedtest1.st-tel.net:8080",
        "d": 1896.6996055651487,
        "latency": 50.377,
    },
    "timestamp": "2021-03-24T00:56:23.048266Z",
    "bytes_sent": 8364032,
    "bytes_received": 26295792,
    "share": None,
    "client": {
        "ip": "66.220.13.38",
        "lat": "38.3479",
        "lon": "-122.9737",
        "isp": "Hurricane Electric",
        "isprating": "3.7",
        "rating": "0",
        "ispdlavg": "0",
        "ispulavg": "0",
        "loggedin": "0",
        "country": "US",
    },
}


@freeze_time("2021-01-01")
@pytest.mark.asyncio
async def test_speed(mocker) -> None:
    mock_speedtest = mock.MagicMock()
    mock_speedtest.Speedtest.return_value.results.dict.return_value = mock_payload
    mocker.patch("senor_octopus.sources.speed.speedtest", mock_speedtest)

    events = [event async for event in speed()]
    print(events)
    assert events == [
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.speedtest.download",
            "value": 20932433.00558606,
        },
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.speedtest.upload",
            "value": 6068322.252775613,
        },
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.speedtest.ping",
            "value": 50.377,
        },
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.speedtest.server",
            "value": {
                "url": "http://speedtest1.st-tel.net:8080/speedtest/upload.php",
                "lat": "39.3958",
                "lon": "-101.0519",
                "name": "Colby, KS",
                "country": "United States",
                "cc": "US",
                "sponsor": "S&T Communications",
                "id": "3434",
                "host": "speedtest1.st-tel.net:8080",
                "d": 1896.6996055651487,
                "latency": 50.377,
            },
        },
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.speedtest.timestamp",
            "value": "2021-03-24T00:56:23.048266Z",
        },
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.speedtest.bytes_sent",
            "value": 8364032,
        },
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.speedtest.bytes_received",
            "value": 26295792,
        },
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.speedtest.share",
            "value": None,
        },
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.speedtest.client",
            "value": {
                "ip": "66.220.13.38",
                "lat": "38.3479",
                "lon": "-122.9737",
                "isp": "Hurricane Electric",
                "isprating": "3.7",
                "rating": "0",
                "ispdlavg": "0",
                "ispulavg": "0",
                "loggedin": "0",
                "country": "US",
            },
        },
    ]



import asyncio
from asyncio.futures import Future
from io import StringIO
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple
from typing import Union

import asciidag.graph
import asciidag.node
from senor_octopus.graph import Node
from senor_octopus.graph import Source
from senor_octopus.types import Event
from senor_octopus.types import Stream


def flatten(
    obj: Dict[str, Any],
    prefix: str = "",
    sep: str = ".",
) -> Dict[str, Union[str, int, float]]:
    items: List[Tuple[str, Union[str, int, float]]] = []
    for key, value in obj.items():
        new_key = sep.join((prefix, key)) if prefix else key
        if isinstance(value, dict):
            items.extend(flatten(value, new_key, sep=sep).items())
        else:
            items.append((new_key, value))
    return dict(items)


def render_dag(dag: Set[Source], **kwargs: Any) -> str:
    out = StringIO()
    graph = asciidag.graph.Graph(out, **kwargs)
    asciidag_nodes: Dict[str, asciidag.node.Node] = {}
    tips = sorted(
        [build_asciidag(node, asciidag_nodes) for node in dag],
        key=lambda n: n.item,
    )

    graph.show_nodes(tips)
    out.seek(0)
    return out.getvalue()


def build_asciidag(
    node: Node,
    asciidag_nodes: Dict[str, asciidag.node.Node],
) -> asciidag.node.Node:
    if node.name in asciidag_nodes:
        asciidag_node = asciidag_nodes[node.name]
    else:
        asciidag_node = asciidag.node.Node(node.name)
        asciidag_nodes[node.name] = asciidag_node

    asciidag_node.parents = sorted(
        [build_asciidag(child, asciidag_nodes) for child in node.next],
        key=lambda n: n.item,
    )

    return asciidag_node


async def merge_streams(*streams: Stream) -> Stream:
    streams_next: Dict[
        Stream,
        Optional[Future[Event]],
    ] = {stream: None for stream in streams}
    stream_map: Dict[Future[Event], Stream] = {}
    while streams_next:
        for stream, next_ in streams_next.items():
            if next_ is None:
                future = asyncio.ensure_future(stream.__anext__())
                stream_map[future] = stream
                streams_next[stream] = future

        done, pending = await asyncio.wait(
            {stream for stream in streams_next.values() if stream},
            return_when=asyncio.FIRST_COMPLETED,
        )
        for future in done:
            # clear message in each completed stream
            stream = stream_map[future]
            streams_next[stream] = None

            try:
                event = future.result()
            except StopAsyncIteration:
                del streams_next[stream]
                continue
            yield event



from datetime import datetime

import pytest
from senor_octopus.sinks.tuya import tuya
from senor_octopus.types import Stream


async def stream() -> Stream:
    for value in {"on", "off", "invalid"}:
        yield {
            "timestamp": datetime(2021, 1, 1),
            "name": "turn",
            "value": value,
        }


@pytest.mark.asyncio
async def test_tuya(mocker) -> None:
    TuyaApi = mocker.patch("senor_octopus.sinks.tuya.TuyaApi")
    api = TuyaApi.return_value
    mock_device = mocker.MagicMock()
    mock_device.name.return_value = "My switch"
    api.get_all_devices.return_value = [mock_device]

    mock_logger = mocker.patch("senor_octopus.sinks.tuya._logger")

    await tuya(stream(), "My switch", "user@example.com", "XXX", "1", "smart_life")

    mock_device.turn_on.assert_called()
    mock_device.turn_off.assert_called()
    mock_logger.warning.assert_called_with("Unknown value: %s", "invalid")


@pytest.mark.asyncio
async def test_tuya_invalid_device(mocker) -> None:
    TuyaApi = mocker.patch("senor_octopus.sinks.tuya.TuyaApi")
    api = TuyaApi.return_value
    mock_device = mocker.MagicMock()
    mock_device.name.return_value = "My switch"
    api.get_all_devices.return_value = [mock_device]

    mock_logger = mocker.patch("senor_octopus.sinks.tuya._logger")

    await tuya(stream(), "My new switch", "user@example.com", "XXX", "1", "smart_life")

    mock_logger.error.assert_called_with(
        'Device "%s" not found. Available devices: %s',
        "My new switch",
        '"My switch"',
    )
